# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_ansible.tests.unit import base


class MockedConfigParser(mock.Mock):
    def __init__(self, conffile, sections):
        super(MockedConfigParser, self).__init__()
        self.sections = sections

    def parse(self):
        self.sections.update({'ansible:testhost': {}})


class TestBuildAnsibleInventory(base.NetworkingAnsibleTestCase):

    def test_build_ansible_inventory_empty_hosts(self):
        self.assertEqual(self.empty_inventory,
                         self.ansconfig.build_ansible_inventory())

    @mock.patch('networking_ansible.config.LOG')
    @mock.patch('networking_ansible.config.cfg.ConfigParser')
    def test_build_ansible_inventory_parser_error(self, mock_parser, mock_log):
        mock_parser().parse.side_effect = IOError()
        self.assertEqual(self.empty_inventory,
                         self.ansconfig.build_ansible_inventory())
        mock_log.error.assert_called()

    @mock.patch('networking_ansible.config.cfg.ConfigParser',
                MockedConfigParser)
    def test_build_ansible_inventory_w_hosts(self):
        self.assertEqual(self.inventory,
                         self.ansconfig.build_ansible_inventory())



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ansible_runner
from oslo_log import log as logging

from networking_ansible import exceptions

LOG = logging.getLogger(__name__)


class AnsibleNetworking(object):
    """Object to invoke ansible_runner to call Ansible Networking

    Hold inventory and provide an interface for calling
    roles in Ansible Networking to manipulate switch configuration
    """

    def __init__(self, inventory):
        self.inventory = inventory

    def _run_task(self, task, host_name, segmentation_id, switch_port=None):
        """Run a task.

        :param task: name of task in openstack-ml2 ansible role
        :param host_name: name of a host defined in ml2 conf ini files
        :param segmentation_id: vlan id of the network
        :param switch_port: port name on the switch (optional)

        See etc/ansible/roles/openstack-ml2/README.md for an exmaple playbook
        """
        # TODO(radez): This is hard coded for juniper because
        #       that was the initial vm we got running.
        #       need to find out how to do this cross vendor
        if segmentation_id == '1':
            segmentation_name = 'default'
        else:
            segmentation_name = 'vlan{}'.format(segmentation_id)

        # build out the ansible playbook
        playbook = [{
            'name': 'Openstack networking-ansible playbook',
            'hosts': host_name,
            'gather_facts': 'no',  # no need to gather facts every run
            'tasks': [{
                'name': 'do {}'.format(task),
                'import_role': {
                    'name': 'openstack-ml2',
                    'tasks_from': task,
                },
                'vars': {
                    'segmentation_name': segmentation_name,
                    'segmentation_id': segmentation_id,
                }
            }]
        }]
        if switch_port:
            playbook[0]['tasks'][0]['vars']['port_name'] = switch_port
            playbook[0]['tasks'][0]['vars']['port_description'] = switch_port

        # TODO(radez) should we pass ident?
        result = ansible_runner.run(playbook=playbook,
                                    inventory=self.inventory,
                                    settings={'pexpect_use_poll': False})
        failures = result.stats['failures']
        if failures:
            raise exceptions.AnsibleRunnerException(' '.join(result.stdout))
        return result

    def create_network(self, host_name, segmentation_id):
        self._run_task('create_network', host_name, segmentation_id)

    def delete_network(self, host_name, segmentation_id):
        self._run_task('delete_network', host_name, segmentation_id)

    def vlan_access_port(self, assign_remove, port, network):
        """Assign an access port to a vlan.

        If the configuration required to unplug the port is not present
        (e.g. local link information), the port will not be unplugged and no
        exception will be raised.

        :param assign_remove: 'assign' or 'remove'
        :param port: The port to unplug
        :param network: The network from which to unplug the port
        """
        task = {'assign': 'update_port',
                'remove': 'delete_port'}
        debug_msg = {'assign': 'Plugging in port {switch_port} on '
                               '{switch_name} to vlan: {segmentation_id}',
                     'remove': 'Unplugging port {switch_port} on '
                               '{switch_name} from vlan: {segmentation_id}'}
        info_msg = {'assign': 'Port {neutron_port} has been plugged into '
                              'network {net_id} on device {switch_name}',
                    'remove': 'Port {neutron_port} has been unplugged from '
                              'network {net_id} on device {switch_name}'}
        error_msg = {'assign': 'Failed to unplug port {neutron_port} on '
                               'device: {switch_name} from network {net_id} '
                               'reason: {exc}',
                     'remove': 'Failed to plug in port {neutron_port} on '
                               'device: {switch_name} from network {net_id} '
                               'reason: {exc}'}

        # If segmentation ID is None, set vlan 1
        segmentation_id = network['provider:segmentation_id'] or '1'

        local_link_info = port['binding:profile'].get('local_link_information')
        if not local_link_info:
            return
        switch_name = local_link_info[0].get('switch_info')
        switch_port = local_link_info[0].get('port_id')
        try:
            LOG.debug(debug_msg[assign_remove].format(
                switch_port=switch_port,
                switch_name=switch_name,
                segmentation_id=segmentation_id))
            self._run_task(task[assign_remove], switch_name,
                           segmentation_id, switch_port)
            LOG.info(info_msg[assign_remove].format(
                neutron_port=port['id'],
                net_id=network['id'],
                switch_name=switch_name))
        except Exception as e:
            LOG.error(error_msg[assign_remove].format(
                neutron_port=port['id'],
                net_id=network['id'],
                switch_name=switch_name,
                exc=e))
            raise



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

service_option = cfg.BoolOpt(
    'netansible',
    default=False,
    help='Whether or not Networking Ansbile mech driver is expected to be '
         'available.')

ovs_opt_group = cfg.OptGroup(
    name='net_ansible_openvswitch',
    title='Settings for OVS provider.',
    help='Options group for OVS Ansible provider.')

OVS_GROUP = [
    cfg.StrOpt(
        'ovsdb_connection',
        default='tcp:127.0.0.1:6640',
        help='Connection string to OVSDB.'),
    cfg.StrOpt(
        'switch_name',
        default='ovs_test',
        help='Name of the switch as presented in the networking ansible '
             'config file.'),
    cfg.StrOpt(
        'bridge_name',
        default='net-ans-br',
        help='Name of the bridge which port is attached to.'),
    cfg.StrOpt(
        'port_name',
        default='net-ans-p0',
        help='Name of the port that is being tested'),
]



# Copyright (c) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def build_ansible_inventory():
    """Get inventory list from config files

    returns python dict representing ansible inventory
    according to ansible inventory file yaml definition
    http://docs.ansible.com/ansible/latest/user_guide/intro_inventory.html
    """
    # TODO(radez): consider take advantage of ansible inventory grouping

    driver_tag = 'ansible:'
    inventory = {}

    for conffile in CONF.config_file:
        # parse each config file
        sections = {}
        parser = cfg.ConfigParser(conffile, sections)
        try:
            parser.parse()
        except IOError as e:
            LOG.error(str(e))

        # filter out sections that begin with the driver's tag
        hosts = {k: v for k, v in sections.items()
                 if k.startswith(driver_tag)}

        # munge the oslo_config data removing the device tag and
        # turning lists with single item strings into strings
        for host in hosts:
            dev_id = host.partition(driver_tag)[2]
            dev_cfg = {k: v[0] for k, v in hosts[host].items()}
            inventory[dev_id] = dev_cfg

    LOG.info('Ansible Host List: %s', ', '.join(inventory))
    return {'all': {'hosts': inventory}}



# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from ovs.db import idl
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.schema.open_vswitch import impl_idl

_idl = None


def idl_factory(conn, schema):
    helper = idlutils.get_schema_helper(conn, schema)
    helper.register_all()

    return idl.Idl(conn, helper)


def get_idl_singleton():
    global _idl

    conn = cfg.CONF.net_ansible_openvswitch.ovsdb_connection
    schema = "Open_vSwitch"

    if _idl is None:
        _connection = connection.Connection(
            idl=idl_factory(conn, schema),
            timeout=10)
        _idl = impl_idl.OvsdbIdl(_connection)

    return _idl



# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from tempest.lib import decorators
from tempest.lib import exceptions

from net_ansible_tempest.tests import base
from net_ansible_tempest.tests import utils


class TestWithOvs(base.NetAnsibleAdminBaseTest):

    @classmethod
    def skip_checks(cls):
        super(TestWithOvs, cls).skip_checks()
        if not cfg.CONF.service_available.netansible:
            raise cls.skipException("networking ansible is not enabled")

    def setUp(self):
        super(TestWithOvs, self).setUp()
        self.ovs = utils.get_idl_singleton()
        self.network = self.create_network()

    def cleanup_port(self, port_id):
        """Remove Neutron port and skip NotFound exceptions."""
        try:
            self.admin_ports_client.delete_port(port_id)
        except exceptions.NotFound:
            pass

    @property
    def ovs_bridge_name(self):
        return cfg.CONF.net_ansible_openvswitch.bridge_name

    @property
    def ovs_bridge_mac(self):
        return self.ovs.db_get(
            'Interface', self.ovs_bridge_name, 'mac_in_use').execute()

    @property
    def switch_name(self):
        return cfg.CONF.net_ansible_openvswitch.switch_name

    @property
    def ovs_port_name(self):
        return cfg.CONF.net_ansible_openvswitch.port_name

    def get_network_segmentation_id(self, network_id):
        return self.admin_networks_client.show_network(
            network_id)['network']['provider:segmentation_id']

    def create_port(self, network_id):
        port = self.admin_ports_client.create_port(
            network_id=network_id, name=self.ovs_port_name)['port']
        self.addCleanup(self.cleanup_port, port['id'])

        host = self.os_admin.hypervisor_client.list_hypervisors(
            )['hypervisors'][0]['hypervisor_hostname']

        llc = [{'switch_info': self.switch_name,
                'switch_id': self.ovs_bridge_mac,
                'port_id': self.ovs_port_name}]

        update_args = {
            'device_owner': 'baremetal:none',
            'device_id': 'fake-instance-uuid',
            'admin_state_up': True,
            'binding:vnic_type': 'baremetal',
            'binding:host_id': host,
            'binding:profile': {
                'local_link_information': llc
            }
        }
        self.admin_ports_client.update_port(
            port['id'],
            **update_args
        )
        return port

    @decorators.idempotent_id('40b81fe4-1e9c-4f10-a808-c23f85aea5e2')
    def test_update_port(self):
        port = self.create_port(self.network['id'])
        current_tag = self.ovs.db_get(
            'Port', self.ovs_port_name, 'tag').execute()
        network_segmentation_id = self.get_network_segmentation_id(
            self.network['id'])
        self.assertEqual(network_segmentation_id, current_tag)

        self.cleanup_port(port['id'])
        current_tag = self.ovs.db_get(
            'Port', self.ovs_port_name, 'tag').execute()
        self.assertEqual([], current_tag)
